


"""
   Builds a Decision Tree using entropy as the impurity measure.

   This script builds a decision tree classifier to predict if the client will
   subscribe a term deposit (variable y). It uses entropy as the impurity measure.
   Uses the Bank Marketing Data Set [1]. The dataset used is the Bank additional data Full dataset. 
   Information about the features/variables can be found in [2].

   References:
      1) Bank Marketing Data Set, https://archive.ics.uci.edu/ml/datasets/Bank+Marketing
      2) File: bank-additional-names.csv

   Required modules:
      * See file requirements.txt 




   v1.0/mmt/Oct 2022
"""






import sys


# Data processing related 
import pandas as pd
import numpy as np
from sklearn.model_selection import train_test_split
from sklearn.tree import DecisionTreeClassifier
from sklearn import preprocessing 
from sklearn.metrics import accuracy_score
from sklearn.metrics import confusion_matrix
from sklearn import tree

# For visualizations
from matplotlib import pyplot as plt



# File name containg data used to build the
# decision tree.
BANK_MARKETING_DATA_FILE = "bank-additional-full.csv"


# Dimensions of figure visualizing the decision tree.
# In inches.
# Note: 1 inch = 2.54 cm
FIGURE_WIDTH  = 20
FIGURE_HEIGHT = 10 




def displayFeatures(df):
    """Displays features and their data type
       
       Parameters:
          df -- data frame. The data frame containing the features that should be presented..
          return -- None.
    """
    for i, c in enumerate(df.columns):
        print('\t',(i+1), ') ', c, ' : ', df.dtypes[c], sep='')





   
    

# Prelude/intro

print("")
print("")
print('*'*92)
print('*')
print('* Decision Trees')
print('*')
print('* Dataset: "Bank Marketing Data Set" available from')
print('*           https://archive.ics.uci.edu/ml/datasets/Bank+Marketing')
print('* Local dataset file: bank-additional-full.csv')      
print('*')
print('*'*92)
print("")



   

# Read the data from the csv file.
# The csv file has a header with the variable names. Separator is the character ;. Also we
# add the argument quotechar="\"" because as you'll see in the file, strings values are enclosed in
# double quotes and we don't want these quotes to be part of the value.

print("Reading dataset from file bank-additional-full.csv....")
bankData = pd.read_csv(BANK_MARKETING_DATA_FILE, sep=";", header=0, quotechar="\"")
print('\tNumber of rows:', bankData.shape[0])
print('\tNumber of columns/features:', bankData.shape[1])
print("Variables of DataFrame bankData:")
displayFeatures(bankData)



     


#########################################################################################################
# Preprocessing
#########################################################################################################



# We remove some variables as these are not relevant 
# for the analysis. Variable 'duration' MUST be removed as it reprepesents
# the duration of the call which is not known beforehand. 
bankData = bankData.loc[:, ~bankData.columns.isin( ['month', 'day_of_week', 'duration'] ) ]
print("Variables of DataFrame bankData after removal:")
displayFeatures(bankData)


#sys.exit(-2)

    
# First, we preprocess the data in order to tranform all categorical (independent)
# variables into a set of so-called dummy variables. In particular, what we do
# here is a One-hot encoding for each categorical variable. One-hot encoding is one
# way of handling such values.
#
# During One-hot encoding, a categorical variable (taking discrete values) is transformed
# into a number of dummy variables, each one of which  represents a value of the categorical
# variable taking binary values i.e. 0 or 1. A value of 1 in one such variable means that
# the categorical variable has this value, while a value of 0 signifies that it does not take
# that value. Hence during One-hot encoding, a categorical variable is transformed into a series
# of 0s and 1 specifying which value the variable takes.
# Because only one variable will have the value 1 and all the others will have the value 0,
# it's called "One-hot".
#
# An example: Let's say that a dataset has a variable named "color" and within that dataset
# it gets one of 5 possible values: green, blue, red, white, black.
# One-hot encoding of the categorical variable color will create 5 new variables - one for each
# VALUE of the variable color- and each of these 5 variable take only binary values i.e. 0 or 1.
# These new variables will be
# green
# blue
# red
# white
# black
# So, if for some row, the value of the variable color is "blue", in terms of the new variables,
# this value will be represented as
#  green | blue | red | white | black
#    0       1     0      0       0
#
# If another row has the value "white" for variable color, this will be represented as
#  green | blue | red | white | black
#    0       0     0      1       0
#
# Now, why are we doing One-hot encoding of all categorical independent variables in our dataset?
# It's because Decision Trees in sklearn, do not support string categorical values (yet).
# As of today, the implementation of Decision Trees in sklearn only supports integer categorical variables i.e.
# categorical variables must be integers such as 0 or 1. Hence we transform all categorical variables
# with string values in our dataset into sets of 0/1 values.
# Due to One-Hot encoding, the number of variables in our dataset will increase. After a categorical
# variable has been One-hot encoded, the initial categorical variable is not needed and can be
# removed from the dataset.
#
# NOTE: One-Hot encoding is one way of doing this. Another way is to use sklearn's LabelEncoder .

    

print("\n\nOne-Hot encoding independent variables in dataset\n")


# First, get a list of the names of variables in our dataset that are categorical.
# We to this using the .select_dtypes().columns method giving as argument [object].
# The value [object] is the way a DataFrame in Python signifies that a variable is categorical.
categoricalVariables=bankData.select_dtypes([object]).columns

# Now variable categoricalVariables has the list of names of variables in the DataFrame
# bankData that have categorical variables. We iterate over these variables and
# do a One-hot encoding of each of these during each iteration.
# CAVEAT: We only avoid the class attribute. The class attribute, that takes binary values 'yes', 'no'
# will not be One-hot encoded. We use a different way to encode the class attribute to 0 and 1.
for var in categoricalVariables:
     print("\tOne-Hot-Encoding variable ",var, " .....", sep="", end="")
     # Is this variable our class attribute? If so, DO NOT One-hot encode it i.e. ignore it
     if var == "y":
        print("Ignored") # yes it is the class attribute. Ignore it for now,
                         #as we will handle it differently
        continue    # don't execute the code below. Go on to next attribute
    
     # If we reach this point during the for-loop, the variable var IS NOT the class attribute
     # and catefffgorical. Hence, One-hot encode it.

     # First, make the variable we are going to One-hot encode explicitly categorical
     bankData[var]=pd.Categorical(bankData[var])
     # Now, this does the actual work of One-hot encoding of variable var.
     # varDummies will be a pandas DataFrame with new variables, one for each
     # possible value of variable var and the proper value for each of these variables
     # for the respective entry in the original dataset.
     varDummies = pd.get_dummies(bankData[var], prefix = var)
     # One-hot encoding of variable done.
     # Now, add the new variables to our original dataset.
     bankData = pd.concat([bankData, varDummies], axis=1)
     # And remove the original variable from our dataset. We don't need it anymore
     # since we have One-hot encoded it and added the new variables to the DataFrame.
     bankData=bankData.drop([var], axis=1)
     print("Done")
     # Finished. Go to next categorical variable in our dataset and do the same.


# If we reach this point, all categorical variables have been One-hot encoded. We print out
# the variables of the DataFrame just to get a look at how many new variables were added


print("\n\tVariables of DataFrame bankData after One-hot encoding:")
print("\t", bankData.columns)

# We left out the class attribute during One-hot encoding.
# Since the class attribute (attribute/variable y in our dataset) already takes two values
# ('yes', 'no') and hence is in essence already a binary variable, we just do different
# encoding of these values: we will transform the value 'no' to the number 0 and the value 'yes'
# to the value 1.
# We do this encoding and add a new column to the end of DataFrame bankData named newY, containing
# these new values of 0 and 1.
# Once we have done this, the original column/variable y is not needed anymore and can be dropped.

bankData['newY'] = (bankData['y'].map( {'no':0, 'yes':1}))



# Drop the column y. Not needed anymore since we have added the column newY
bankData=bankData.drop(['y'], axis=1)
print("\n\nPreprocessing done.")

# Ok we are done with preprocessing the dataset. Now, separate the features that we will use
# for predicting the class attribute, from the values of the class attribute.

# These are our features i.e. the variable values that we will use to predict the class attribute.
# Remember that our class attribute is called newY and is the LAST one in our DataFrame. Hence
# we select the features with :-1 which means all columns except lst one (which is the class attribute).
features = bankData.iloc[:, :-1]

# Get the values of the class variable i.e. the newY variable/column
classVariable = bankData["newY"]

# Split the DataFrame into a training and testing set.
# We use here the special function train_test_split which does the work for us.
# By saying test_size=0.3 this means that 30% of rows in the original dataset will be used for testing.
# Which rows will be selected is determined randomly. The random_state argument specifies exactly
# this (random selection of the testing set).
# The other 70% of the dataset will be used for training the model
trainingSetFeatures, testingSetFeatures, trainingSetClass, testingSetClass = train_test_split(features,
                                                                                              classVariable,
                                                                                              test_size=0.3,
                                                                                              random_state = 100)



#########################################################################################################
# Exploratory Data Analysis (EDA)
#########################################################################################################

# See the distribution of the dependent variable. Check for balance.
bankData['newY'].value_counts().plot(kind='barh', color='red')

# Display some boxplots
#plt.figure()
fig, axs = plt.subplots(1, 2, figsize=(15, 5))
axs[0].boxplot(bankData['euribor3m'])
axs[0].set_title('Euribor')
axs[0].set_xlabel('euribor3m')
axs[0].set_ylabel('Value')


axs[1].boxplot(bankData['nr.employed'])
axs[1].set_title('Number employees')
axs[1].set_xlabel('nr.employed')
axs[1].set_ylabel('Value')

plt.tight_layout()
plt.show()







#########################################################################################################
# Building Dection tree
#########################################################################################################


# We are now ready to train our Decision Tree (aka the model) with our training set
print("\nTraining the model (decision tree)......", sep='', end='')

# We use entropy as our impurity measure and there set the argument criterion="entropy".
# We could also use the gini measure and in this case we would specify the argument criterion="gini"
termDepositPredictionModel = DecisionTreeClassifier(criterion="entropy")


# Train now the model. The next line will create the Decision Tree using the entropy measure
# with which we are able to predict the value of the class variable newY.
model_fit = termDepositPredictionModel.fit(trainingSetFeatures, trainingSetClass)

print('done.')







#########################################################################################################
# Visualizing decision tree
#########################################################################################################


    
# Visualize/display the estimated/grown/built Decision Tree.
# We'll use two different ways to do so


# 1st way: Visualizing the Decision Tree using a text representation
textRepresentation = tree.export_text(termDepositPredictionModel)
print(textRepresentation)



# 2nd way: Displaying the Decision Tree in a more colorful way
#          including some important information related to probabilities of data falling
#          in each class.
#
# NOTE: We display only the decision tree at depth 3, because the full tree is too large.
#       Change max_depth argument below
#       to change display depth. Leave max_depth parameter out to visualize entire tree
#       (CAVEAT! This may take a long time. You have been warned).


# Prepare the dimensions of the figure upon which
# the decision tree will be visualized. The decision tree may be large
# and width, height may be needed to be configured properly to
# show the decision tree in an acceptable way.
plt.figure(figsize=(FIGURE_WIDTH, FIGURE_HEIGHT))


# Plot the tree as specified by the arguments:
# model : the Decicion Tree i.e. the actual model, that has been generated/grown
# max_depth : the depth at which to display the Decision Tree. We set this to 3 since
#             the full Decision Tree is very large and will take time to be visualized.
#             Leave this parameter out, to visualize entire tree
# propotion : show number of observations in training set reaching that node as percentage
# filled : use color of nodes to indicate majority of class.

_ = tree.plot_tree(termDepositPredictionModel,
                   feature_names=trainingSetFeatures.columns,
                   max_depth = 3,
                   proportion = True,
                   filled = True)

 # Set the visualization title
plt.title("Decision tree generated from training data (visualized only at depth=3, not fully)",
           fontsize=16,  color= 'red', fontweight='bold',loc='center')


# Everything seems ok. Display the decision tree now
plt.show()




#########################################################################################################
# Decision tree model evaluation
#########################################################################################################

    

# Decision tree has been built and visualized. Now use the testing set in order to
# assess the accuracy of the Decision Tree.
# We use the .predict() method of the model which, when given as
# input the values of the features will return an array with values 0 and 1, each one of these
# is the predicted value for the respective row in the testing set. This means the the first value
# in the array returned is the predicted value for attribute newY for the data in the first row of
# the testing set, the second value in the array returned is the predicted value for attribute
# newY for the second row in the testing set etc.
print("\nUsing testing set to predict class attribute......", sep='', end='')
predictions = termDepositPredictionModel.predict(testingSetFeatures)
print("done.")



# Since we have the predicted values for our testing set (predictions) and the real values for the
# testing set (testingSetClass) we can assess how good our model is in predicting the
# variable newY by calculating the confusion matrix.
# The confusion matrix will tell us how many correct predictions the model made and how
# many incorrect predictions. This in turn can be used to calculate the accuracy of the
# model (see below).
# Confusion matrix is transformed into a data frame to make display cleaner.
print("\nCalculating confusion matrix on the testing set......", sep='', end='')
testSetConfusionMatrix = pd.DataFrame(confusion_matrix(testingSetClass, predictions))

print("done.")
# Print/show the confusion matrix
print("Confusion matrix:")
print(testSetConfusionMatrix)

# Next, based on the values in the confusion matrix, the accuracy
# of the model will be calculated using the accuracy_score function in sklearn.metrics.
# A rule of thumb: an accuracy of > 70% means the model can be considered useful.
print (f"\nModel's predictive accuracy is: {(accuracy_score(testingSetClass, predictions)*100):.4}%" )
    




#
# Displaying the important features
#




# Decision trees have also explanatory power.
# E.g. what are the most important features? Importance here is defined as how good each
# attribute is in splitting the data in low-impurity subsets.
# The numbers you'll see are the gain in entropy induced by each feature. I.e. how much improvement
# of impurity each feature achiueves (the more, the better) 
for f, imp in zip(testingSetFeatures.columns, termDepositPredictionModel.feature_importances_):
    print(f,":", imp)

# Display importance as barchart
plt.barh( y=testingSetFeatures.columns,  width =termDepositPredictionModel.feature_importances_, color ='maroon')

plt.xlabel("Features")
plt.ylabel("Importance (improvement of entropy)")
plt.title("Feature importance")
plt.show()

